import numpy as np
import datetime as dt
import plotly.express as px

from .utils import get_correctly_formatted_particle_dist_list
from .anisotropic_MAIRE_engine.spectralCalculations.rigiditySpectrum import DLRmodelSpectrum, MishevModifiedPowerLawSpectrum, MishevModifiedPowerLawSpectrumSplit
from .anisotropic_MAIRE_engine.spectralCalculations.pitchAngleDistribution import gaussianBeeckPitchAngleDistribution, isotropicPitchAngleDistribution, gaussianPitchAngleDistribution

from .anisotropic_MAIRE_engine.maireSengine import generalEngineInstance, default_array_of_lats_and_longs

default_altitudes_in_kft = [0,10,20] + [i for i in range(25, 61 + 1, 3)]

def run_from_spectra(
        proton_rigidity_spectrum=None,
        alpha_rigidity_spectrum=None,
        reference_pitch_angle_latitude=0.0, #None,
        reference_pitch_angle_longitude=45.0, #None,
        proton_pitch_angle_distribution=isotropicPitchAngleDistribution(),
        alpha_pitch_angle_distribution=isotropicPitchAngleDistribution(),
        altitudes_in_kft=default_altitudes_in_kft,
        altitudes_in_km=None,
        Kp_index=None,
        date_and_time=dt.datetime.now(),
        array_of_lats_and_longs=default_array_of_lats_and_longs,
        cache_magnetocosmics_run=True,
        generate_NM_count_rates=True,
        **mag_cos_kwargs,
):
    
    if Kp_index is None:
        raise Exception("Error: no Kp index specified!")
    
    if (altitudes_in_km is not None) and (altitudes_in_kft is not default_altitudes_in_kft):
        raise Exception("Error: only one of altitudes_in_km and altitudes_in_kft should be supplied!")
    
    if altitudes_in_km is None:
        altitudes_in_km = np.array(altitudes_in_kft) * 0.3048

    if (proton_rigidity_spectrum is None) and (alpha_rigidity_spectrum is None):
        raise Exception("Error: either a proton rigidity spectrum or an alpha rigidity spectrum must be specified!")
    
    list_of_particle_distributions = get_correctly_formatted_particle_dist_list(proton_rigidity_spectrum, 
                                                                                alpha_rigidity_spectrum, 
                                                                                reference_pitch_angle_latitude, 
                                                                                reference_pitch_angle_longitude, 
                                                                                proton_pitch_angle_distribution, 
                                                                                alpha_pitch_angle_distribution)

    
    engine_to_run = generalEngineInstance(list_of_particle_distributions,
                                          list_of_altitudes_km=altitudes_in_km,
                                          Kp_index=Kp_index,
                                          date_and_time=date_and_time,
                                          reference_latitude=reference_pitch_angle_latitude,
                                          reference_longitude=reference_pitch_angle_longitude,
                                          array_of_lats_and_longs=array_of_lats_and_longs,
                                          cache_magnetocosmics_runs=cache_magnetocosmics_run,
                                          generate_NM_count_rates=generate_NM_count_rates)
    
    output_dose_rate_DF = engine_to_run.getAsymptoticDirsAndRun(**mag_cos_kwargs)

    print("Success!")

    return output_dose_rate_DF

def run_from_power_law_gaussian_distribution(J0, gamma, deltaGamma, sigma, 
                                             reference_pitch_angle_latitude, reference_pitch_angle_longitude, 
                                             Kp_index,date_and_time,
                                             use_split_spectrum=False,
                                             **kwargs):
    
    if use_split_spectrum == True:
        spec_to_use = MishevModifiedPowerLawSpectrumSplit
    else:
        spec_to_use = MishevModifiedPowerLawSpectrum
        

    output_dose_rate_DF = run_from_spectra(
        proton_rigidity_spectrum=spec_to_use(J0, gamma, deltaGamma),
        reference_pitch_angle_latitude=reference_pitch_angle_latitude,
        reference_pitch_angle_longitude=reference_pitch_angle_longitude,
        proton_pitch_angle_distribution=gaussianPitchAngleDistribution(normFactor=1,sigma=sigma),
        Kp_index=Kp_index,date_and_time=date_and_time,
        **kwargs,
    )

    return output_dose_rate_DF

def run_from_double_power_law_gaussian_distribution(J0, gamma, deltaGamma, sigma_1, sigma_2,
                                                    B, alpha_prime,
                                             reference_pitch_angle_latitude, reference_pitch_angle_longitude, 
                                             Kp_index,date_and_time,
                                             use_split_spectrum=False,
                                             **kwargs):
    
    if use_split_spectrum == True:
        spec_to_use = MishevModifiedPowerLawSpectrumSplit
    else:
        spec_to_use = lambda J0,gamma,deltaGamma:MishevModifiedPowerLawSpectrum(J0,gamma,deltaGamma, lowerLimit=0.814529,upperLimit=21.084584)
        

    output_dose_rate_DF = run_from_spectra(
        proton_rigidity_spectrum=spec_to_use(J0, gamma, deltaGamma),
        reference_pitch_angle_latitude=reference_pitch_angle_latitude,
        reference_pitch_angle_longitude=reference_pitch_angle_longitude,
        proton_pitch_angle_distribution=gaussianPitchAngleDistribution(normFactor=1,sigma=sigma_1) + (B * gaussianPitchAngleDistribution(normFactor=1,sigma=sigma_2,alpha=alpha_prime)),
        Kp_index=Kp_index,date_and_time=date_and_time,
        **kwargs,
    )

    return output_dose_rate_DF

def run_from_power_law_Beeck_gaussian_distribution(J0, gamma, deltaGamma, A, B, 
                                             reference_pitch_angle_latitude, reference_pitch_angle_longitude, 
                                             Kp_index,date_and_time,
                                             use_split_spectrum=False,
                                             **kwargs):
    
    if use_split_spectrum == True:
        spec_to_use = MishevModifiedPowerLawSpectrum
    else:
        spec_to_use = MishevModifiedPowerLawSpectrumSplit

    output_dose_rate_DF = run_from_spectra(
        proton_rigidity_spectrum=spec_to_use(J0, gamma, deltaGamma),
        reference_pitch_angle_latitude=reference_pitch_angle_latitude,
        reference_pitch_angle_longitude=reference_pitch_angle_longitude,
        proton_pitch_angle_distribution=gaussianBeeckPitchAngleDistribution(normFactor=1,A=A,B=B),
        Kp_index=Kp_index,date_and_time=date_and_time,
        **kwargs,
    )

    return output_dose_rate_DF

def run_from_DLR_cosmic_ray_model(OULU_count_rate_in_seconds=None,
                                      W_parameter=None,
                                      Kp_index=None,date_and_time=None,
                                      **kwargs):
    
    if (W_parameter is None) and (OULU_count_rate_in_seconds is None):
        print("As no OULU count rates or W parameters were specified, the count rate of OULU in the past will be determined using the supplied date and time for the purposes of calculating the incoming spectra.")
        DLR_model_date_and_time = date_and_time
    else:
        DLR_model_date_and_time = None

    output_dose_rate_DF = run_from_spectra(
        proton_rigidity_spectrum=DLRmodelSpectrum(atomicNumber=1, date_and_time=DLR_model_date_and_time, OULUcountRateInSeconds=OULU_count_rate_in_seconds, W_parameter=W_parameter),
        alpha_rigidity_spectrum=DLRmodelSpectrum(atomicNumber=2, date_and_time=DLR_model_date_and_time, OULUcountRateInSeconds=OULU_count_rate_in_seconds, W_parameter=W_parameter),
        Kp_index=Kp_index,date_and_time=date_and_time,
        **kwargs,
    )

    return output_dose_rate_DF

def create_single_dose_map_plot(DF_to_use,
                              selected_altitude_in_km):

    if selected_altitude_in_km is not None:
        DF_to_use = DF_to_use[round(DF_to_use["altitude (km)"],4) == selected_altitude_in_km]

    if len(DF_to_use) == 0:
        raise Exception("Error: specified altitude in kilometers did not match any of the altitudes in kilometers in the inputted DataFrame.")

    doseRateMap = px.scatter(DF_to_use, x="longitude",y="latitude",color="adose",
                            symbol_sequence=["square"],
                            range_y=[-90,90],
                            range_x=[0,360])
    
    doseRateMap.update_traces(marker={'size': 10})
    doseRateMap.update_layout(yaxis=dict(scaleanchor="x", scaleratio=1, range=[-90,90]))
    doseRateMap.update_xaxes(range=[0,360])
    doseRateMap.update_yaxes(range=[-90,90])
    doseRateMap.update_layout(xaxis_scaleanchor="y")

    doseRateMap.update_layout(autosize=False,
                            width=800,
                            height=600)

    doseRateMap.show()

    return doseRateMap

import seaborn as sns
import geopandas
import matplotlib.pyplot as plt

def create_single_dose_map_plot_plt(heatmap_DF_to_Plot, hue_range = (2,7), heatmap_s = 63,edgecolor='white',
                    dose_type = "adose"):

    ############################ creating background world map and dose image
    currentFigure = plt.gcf()

    currentFigure.set_figheight(10)
    currentFigure.set_figwidth(10)

    #heatmap_DF_to_Plot = pd.read_csv(file_path_to_read, delimiter=',')
    heatmap_DF_to_Plot["SEU (Upsets/hr/Gb)"] = heatmap_DF_to_Plot["SEU"] * (60.0 * 60.0) * 1e9
    heatmap_DF_to_Plot["SEL (Latch-ups/hr/device)"] = heatmap_DF_to_Plot["SEL"] * (60.0 * 60.0)
    heatmap_DF_to_Plot["longitudeTranslated"] = heatmap_DF_to_Plot["longitude"].apply(lambda x:x-360.0 if x > 180.0 else x)

    scatterPlotAxis = sns.scatterplot(data=heatmap_DF_to_Plot,x="longitudeTranslated",y="latitude",
                    hue=dose_type, hue_norm=hue_range, palette="crest",
                    zorder=10,
                    marker="s",s=heatmap_s,edgecolor=edgecolor,
                    legend="brief",
                    )#ax=axToPlotOn)

    world = geopandas.read_file(geopandas.datasets.get_path('naturalearth_lowres'))
    world.plot(color="None",edgecolor="black",lw=1.5,ax=scatterPlotAxis,zorder=20)

    ####################################################################

    plt.ylim([-90,90])
    plt.xlim([-175,180])
    plt.grid(True)
    plt.xlabel("Longitude (degrees)")
    plt.ylabel("Latitude (degrees)")

    plt.legend(title=dose_type,loc="center left",bbox_to_anchor=(1.1,0.5))

    return plt.gca()


    




from scipy.interpolate import interp1d
import pandas as pd
import scipy
import copy

import pickle as pkl
import numpy as np

from CosRayModifiedISO import CosRayModifiedISO

print("WARNING: currently unknown whether the reported spectral weighting factor should be in terms of energy or rigidityd")

class rigiditySpectrum():

    def __init__(self):
        pass

    def __call__(self, x):
        return self.rigiditySpec(x)
    
    def __add__(self, right):
        summed_spectrum = rigiditySpectrum()
        summed_spectrum.rigiditySpec = lambda x:self.rigiditySpectrum(x) + right.rigiditySpectrum(x)
        return summed_spectrum

class powerLawSpectrum(rigiditySpectrum):

    rigiditySpec = lambda self,x:self.normalisationFactor * (x**self.spectralIndex)

    def __init__(self, normalisationFactor, spectralIndex):

        self.normalisationFactor = normalisationFactor
        self.spectralIndex = spectralIndex

class interpolatedInputFileSpectrum(rigiditySpectrum):

    def __init__(self, inputFileName):

        self.inputFilename = inputFileName
        self.rigiditySpec = self.readSpecFromCSV(self.inputFilename)

    def readSpecFromCSV(self, inputFileName):

        inputDF = pd.read_csv(inputFileName,header=None)
        rigidityList = inputDF[0] # GV
        fluxList = inputDF[1] # p/m2/sr/s/GV
        fluxListcm2 = fluxList / (100**2)

        #rigiditySpec = scipy.interpolate.interp1d(rigidityList,fluxListcm2,kind="linear",fill_value="extrapolate")
        rigiditySpec = scipy.interpolate.interp1d(rigidityList,fluxListcm2,kind="linear",
                                                  fill_value=0.0, bounds_error=False)
        
        return rigiditySpec

class DLRmodelSpectrum(rigiditySpectrum):

    def __init__(self, atomicNumber, date_and_time=None, OULUcountRateInSeconds=None, W_parameter=None):

        if not sum([(date_and_time is not None),(OULUcountRateInSeconds is not None),(W_parameter is not None)]) == 1:
            print("Error: exactly one supplied input out of the date and time, OULU count rate per second or the W paramer, to the DLR model spectrum must be given!")
            raise Exception

        if date_and_time is not None:
            self._generatedSpectrumDF = CosRayModifiedISO.getSpectrumUsingTimestamp(timestamp=date_and_time, atomicNumber=atomicNumber)

        if W_parameter is not None:
            self._generatedSpectrumDF = CosRayModifiedISO.getSpectrumUsingSolarModulation(solarModulationWparameter=W_parameter, atomicNumber=atomicNumber)

        if OULUcountRateInSeconds is not None:
            self._generatedSpectrumDF =CosRayModifiedISO.getSpectrumUsingOULUcountRate(OULUcountRatePerSecond=OULUcountRateInSeconds, atomicNumber=atomicNumber)

        self.rigiditySpec = interp1d(x=self._generatedSpectrumDF["Rigidity (GV/n)"],
                                                        y=self._generatedSpectrumDF["d_Flux / d_R (cm-2 s-1 sr-1 (GV/n)-1)"],
                                                        kind="linear",
                                                        bounds_error=False,
                                                        fill_value = (0.0,0.0))

        with open("DLRsavedSpectrum.pkl","wb") as DLRspecFile:
            pkl.dump(self,DLRspecFile)

class MishevModifiedPowerLawSpectrum(rigiditySpectrum):

    specIndexModification = lambda self,P:self.deltaGamma * (P-1)
    rigiditySpec = lambda self,P:self.J0 * self.step_function(P, self.lowerLimit, self.upperLimit) * (P**(-(self.gamma + self.specIndexModification(P)))) / (100**2) #cm-2 s-1 sr-1 GV-1 : converted from m-2 to cm-2

    def __init__(self,J0,gamma,deltaGamma,lowerLimit=-1 * np.inf,upperLimit=np.inf):

        self.lowerLimit = lowerLimit
        self.upperLimit = upperLimit

        self.J0 = J0 #m-2 s-1 sr-1 GV-1
        self.gamma = gamma
        self.deltaGamma = deltaGamma

        with open("MishevsavedSpectrum.pkl","wb") as MishevspecFile:
            pkl.dump(self,MishevspecFile)

    def step_function(self, rigidity, lowerLimit, upperLimit):
        
        if (rigidity >= lowerLimit) and (rigidity <= upperLimit):
            return 1.0
        else:
            return 0.0

class MishevModifiedPowerLawSpectrumSplit(rigiditySpectrum):

    specIndexModification_high = lambda self,P:self.deltaGamma * (P-1)
    specIndexModification_low = lambda self,P:self.deltaGamma * (P)
    specIndexModification = lambda self,P:self.specIndexModification_high(P) if P > 1.0 else self.specIndexModification_low(P)
    rigiditySpec = lambda self,P:self.J0 * (P**(-(self.gamma + self.specIndexModification(P)))) / (100**2) #cm-2 s-1 sr-1 GV-1 : converted from m-2 to cm-2

    def __init__(self,J0,gamma,deltaGamma):

        self.J0 = J0 #m-2 s-1 sr-1 GV-1
        self.gamma = gamma
        self.deltaGamma = deltaGamma

        with open("MishevsavedSpectrum.pkl","wb") as MishevspecFile:
            pkl.dump(self,MishevspecFile)




import datetime as dt

import numpy as np
from anisotropicMAIRE import anisotropicMAIRE

def test_run_from_spectra():

    anisotropicMAIRE.run_from_spectra(proton_rigidity_spectrum=lambda x:1,
                     alpha_rigidity_spectrum=lambda x:1,
                     Kp_index=6,
                     date_and_time=dt.datetime(year=2000,month=7,day=14,hour=10,minute=45,second=0))
    
def test_run_from_spectra_two_locations():

    anisotropicMAIRE.run_from_spectra(proton_rigidity_spectrum=lambda x:1,
                     alpha_rigidity_spectrum=lambda x:1,
                     Kp_index=6,
                     date_and_time=dt.datetime(year=2000,month=7,day=14,hour=10,minute=45,second=0),
                     altitudes_in_km=np.append(np.array(range(0,13)) * 0.3048,11.28),
                     array_of_lats_and_longs=[[46.2,187.4],[-28.3,-92.7]])
    
def test_run_from_spectra_proton_only():

    anisotropicMAIRE.run_from_spectra(proton_rigidity_spectrum=lambda x:1,
                     Kp_index=6,
                     date_and_time=dt.datetime(year=2000,month=7,day=14,hour=10,minute=45,second=0))
    
def test_Mishev_spec():
    anisotropicMAIRE.run_from_power_law_gaussian_distribution(J0=256_000.0, gamma=3.41, deltaGamma=0.22, sigma=np.sqrt(0.19),
                                         reference_pitch_angle_latitude=-17.0,reference_pitch_angle_longitude=148.0,
                                         Kp_index=3, date_and_time=dt.datetime(2006, 12, 13, 3, 0),
                                         altitudes_in_km=np.append(np.array(range(0,13)) * 0.3048,11.28),
                                         )
    
def test_Mishev_spec_max_asymp_dir():

    array_of_lats_and_longs = np.array([[65.0,25.0],[-35.0,78.0]])
    array_of_zeniths_and_azimuths = np.array([(i,j) for i in np.linspace(0,20,5) for j in np.linspace(0,360,5)])

    outputted_dose = anisotropicMAIRE.run_from_power_law_gaussian_distribution(J0=256_000.0, gamma=3.41, deltaGamma=0.22, sigma=np.sqrt(0.19),
                                         reference_pitch_angle_latitude=-17.0,reference_pitch_angle_longitude=148.0,
                                         Kp_index=3, date_and_time=dt.datetime(2006, 12, 13, 3, 0),
                                         altitudes_in_km=np.append(np.array(range(0,13)) * 0.3048,11.28),
                                         array_of_lats_and_longs=array_of_lats_and_longs,
                                         array_of_zeniths_and_azimuths=array_of_zeniths_and_azimuths,
                                         )
    
    print(outputted_dose)


def test_DLR_spec():
    anisotropicMAIRE.run_from_DLR_cosmic_ray_model(OULU_count_rate_in_seconds=106.0,
                                         Kp_index=3, date_and_time=dt.datetime(2006, 12, 13, 3, 0),
                                         altitudes_in_km=np.append(np.array(range(0,13)) * 0.3048,11.28),
                                         )


import numpy as np

from .pitchAngleDistribution import pitchAngleDistribution
from .rigiditySpectrum import rigiditySpectrum

def jacobian_function_to_use(pitch_angle_in_radians):
        try:
            output_Jacobian_factor = 1/np.sin(2.0 * pitch_angle_in_radians)
        except ZeroDivisionError:
            output_Jacobian_factor =  0.0
        return output_Jacobian_factor

class momentaDistribution():

    def __init__(self, rigidity_spectrum:rigiditySpectrum, pitch_angle_distribution:pitchAngleDistribution):
        self.setRigiditySpectrum(rigidity_spectrum)
        self.setPitchAngleDistribution(pitch_angle_distribution)

    # setter methods

    def setPitchAngleDistribution(self,pitch_angle_distribution:pitchAngleDistribution):
        self.pitch_angle_distribution = pitch_angle_distribution

    def setRigiditySpectrum(self, rigidity_spectrum):
        self.rigidity_spectrum = rigidity_spectrum

    # getter methods

    def getPitchAngleDistribution(self):
        return self.pitch_angle_distribution

    def getRigiditySpectrum(self)->rigiditySpectrum:
        return self.rigidity_spectrum

    # other Methods

    def __call__(self, pitchAngle, rigidity):

        pitch_angle_weighting_factor = self.pitch_angle_distribution(pitchAngle,rigidity) #* jacobian_function_to_use(pitchAngle)
        rigidity_weighting_factor = self.rigidity_spectrum(rigidity)

        full_rigidity_pitch_weighting_factor = pitch_angle_weighting_factor * rigidity_weighting_factor

        return full_rigidity_pitch_weighting_factor


from .particleSpecies import particleSpecies
from .momentaDistribution import momentaDistribution
from .rigiditySpectrum import rigiditySpectrum
from .pitchAngleDistribution import pitchAngleDistribution

class particleDistribution():

    def __init__(self, 
                 particle_name:str, 
                 rigidity_spectrum:rigiditySpectrum, 
                 pitch_angle_distribution:pitchAngleDistribution):

        self.particle_species = particleSpecies(particle_name)
        self.momentum_distribution = momentaDistribution(rigidity_spectrum,
                                                         pitch_angle_distribution)




from .anisotropic_MAIRE_engine.spectralCalculations.particleDistribution import particleDistribution
from .anisotropic_MAIRE_engine.spectralCalculations.pitchAngleDistribution import isotropicPitchAngleDistribution, pitchAngleDistribution


def get_correctly_formatted_particle_dist_list(proton_rigidity_spectrum, 
                                               alpha_rigidity_spectrum, 
                                               reference_pitch_angle_latitude, 
                                               reference_pitch_angle_longitude, 
                                               proton_pitch_angle_distribution, 
                                               alpha_pitch_angle_distribution):
    list_of_particle_distributions = []
    for particle_name, particle_rigidity_spectrum, particle_pitch_angle_distribution in [["proton",proton_rigidity_spectrum,proton_pitch_angle_distribution], 
                                                                                         ["alpha",alpha_rigidity_spectrum,alpha_pitch_angle_distribution]]:
        if not particle_rigidity_spectrum is None:
            if not type(particle_pitch_angle_distribution).__name__ == 'isotropicPitchAngleDistribution':
                if (reference_pitch_angle_latitude is None) or (reference_pitch_angle_longitude is None):
                    raise Exception("Error: a non-isotropic pitch angle distribution was used but reference longitudes and latitudes were not set!")

                particle_pitch_angle_distribution = pitchAngleDistribution(particle_pitch_angle_distribution,
                                                                  reference_pitch_angle_latitude,
                                                                  reference_pitch_angle_longitude)
                
            particle_distribution = particleDistribution(
                                                particle_name,
                                                particle_rigidity_spectrum,
                                                particle_pitch_angle_distribution)
            
            list_of_particle_distributions.append(particle_distribution)
    return list_of_particle_distributions




#!/bin/python3
import numpy as np
import pandas as pd
from tqdm import tqdm

from .singleParticleEngineInstance import singleParticleEngineInstance

tqdm.pandas()


from AsympDirsCalculator import AsympDirsTools

from .utils.AsymptoticDirectionDataframe import generate_asymp_dir_DF

import datetime as dt

from joblib import Memory
MAGCOScachedir = 'cachedMagnetocosmicsRunData'
MAGCOSmemory = Memory(MAGCOScachedir, verbose=0)
default_array_of_lats_and_longs = np.array(np.meshgrid(np.linspace(-90.0, 90.0, 37), np.linspace(0.0, 355.0, 72))).T.reshape(-1, 2)

class generalEngineInstance():

    def __init__(self, 
                 list_of_particle_distributions:list,
                 list_of_altitudes_km:list, 
                 Kp_index:int, 
                 date_and_time:dt.datetime,
                 reference_latitude=0.0, #None,
                 reference_longitude=45.0, #None,
                 array_of_lats_and_longs=default_array_of_lats_and_longs,
                 cache_magnetocosmics_runs=True,
                 generate_NM_count_rates=True):
        
        self.rigiditySpectrumParamDict = {}
        self.pitchAngleDistributionParamDict = {}

        self.list_of_particle_distributions = list_of_particle_distributions
        self.list_of_altitudes_km = list_of_altitudes_km
        self.Kp_index = Kp_index
        self.date_and_time = date_and_time
        self.reference_latitude = reference_latitude
        self.reference_longitude = reference_longitude
        self.array_of_lats_and_longs = array_of_lats_and_longs

        self.cache_magnetocosmics_runs = cache_magnetocosmics_runs
        self.generate_NM_count_rates = generate_NM_count_rates

    def getAsymptoticDirsAndRun(self,**mag_cos_kwargs)->pd.DataFrame:
        
        self.acquireDFofAllAsymptoticDirections(**mag_cos_kwargs)

        fullDoseRateList = []

        for incoming_particle_distribution in self.list_of_particle_distributions:
            singleParticleEngine = singleParticleEngineInstance(incoming_particle_distribution, 
                                                                self.df_of_asymptotic_directions,
                                                                self.list_of_altitudes_km,
                                                                self.generate_NM_count_rates)
            
            doseRateDFforParticleSpecies = singleParticleEngine.runOverSpecifiedAltitudes()

            fullDoseRateList.append(doseRateDFforParticleSpecies)

        summedDoseRateDF = fullDoseRateList[0]
        for doseRateDF in fullDoseRateList[1:]:
            for doseRateName in ["adose","edose","dosee","SEU","SEL"]:
                summedDoseRateDF[doseRateName] += doseRateDF[doseRateName]

        return summedDoseRateDF
    
    def acquireDFofAllAsymptoticDirections(self, **mag_cos_kwargs):
        # run magnetocosmics to get asymptotic directions
        raw_asymp_dir_DF = AsympDirsTools.get_magcos_asymp_dirs(
                                                array_of_lats_and_longs=self.array_of_lats_and_longs,
                                                KpIndex=self.Kp_index,
                                                dateAndTime=self.date_and_time,
                                                cache=self.cache_magnetocosmics_runs,
                                                full_output=True,
                                                **mag_cos_kwargs,
                                                )
        raw_asymp_dir_DF.to_pickle("raw_asymp_dir_DF.pkl")
        self.df_of_asymptotic_directions = generate_asymp_dir_DF(raw_asymp_dir_DF, 
                                                                 self.reference_latitude, 
                                                                 self.reference_longitude, 
                                                                 self.date_and_time,
                                                                 cache = False)
        self.df_of_asymptotic_directions.to_pickle("self_df_of_asymptotic_directions.pkl")










class particleSpecies():

    particle_atomic_number_dict = {"proton":1,"alpha":2}

    def __init__(self, particleName, atomicNumber=None):
        self.particleName = particleName
        if atomicNumber == None:
            atomicNumber = self.particle_atomic_number_dict[particleName]
        
        self.atomicNumber = atomicNumber


import numpy as np
from spacepy.coordinates import Coords as spaceCoords
import pickle as pkl
import copy

class pitchAngleDistribution():

    def __init__(self, 
                 pitch_angle_distribution:callable, 
                 reference_latitude_in_GSM, 
                 reference_longitude_in_GSM):

        self.pitchAngleDistFunction = pitch_angle_distribution

        self.interplanetary_mag_field = spaceCoords([100.0,
                                                     reference_latitude_in_GSM, 
                                                     reference_longitude_in_GSM],
                                                     "GSM","sph")

    def __call__(self,pitchAngle, rigidity):
        #print(f"pitch weighting factor: {self.pitchAngleDistFunction(pitchAngle, rigidity)}")
        return self.pitchAngleDistFunction(pitchAngle, rigidity)
    
    def __add__(self, right):
        summed_dist = copy.deepcopy(self)
        summed_dist.pitchAngleDistFunction = lambda x,y:self.pitchAngleDistFunction(x,y) + right.pitchAngleDistFunction(x,y)
        return summed_dist
    
    def __mul__(self, right:float):
        multiplied_dist = copy.deepcopy(self)
        multiplied_dist.pitchAngleDistFunction = lambda x,y:right * self.pitchAngleDistFunction(x,y)
        return multiplied_dist

    __rmul__ = __mul__

class cosinePitchAngleDistribution(pitchAngleDistribution):

    pitchAngleDistFunction = lambda self,pitchAngle, rigidity:np.abs(0.5*np.sin(2*pitchAngle))

    def __init__(self):
        pass

class isotropicPitchAngleDistribution(pitchAngleDistribution):

    #pitchAngleDistFunction = lambda self,pitchAngle,theta:1/np.pi
    pitchAngleDistFunction = lambda self,pitchAngle,rigidity:1
    #pitchAngleDistFunction = lambda self,pitchAngle,rigidity:np.sin(2.0 * pitchAngle)

    def __init__(self):
        # with open("DLRsavedPAD.pkl","wb") as DLRPADFile:
        #     pkl.dump(self,DLRPADFile)
        pass

class gaussianPitchAngleDistribution(pitchAngleDistribution):

    pitchAngleDistFunction = lambda self,pitchAngle,rigidity:self.normFactor * np.exp(-(pitchAngle - self.alpha)**2/(self.sigma**2))

    def __init__(self,normFactor,sigma, alpha=0.0):
        self.normFactor = normFactor
        self.sigma = sigma
        self.alpha = alpha
        with open("MishevsavedPAD.pkl","wb") as MishevPADFile:
            pkl.dump(self,MishevPADFile)

class gaussianBeeckPitchAngleDistribution(pitchAngleDistribution):

    pitchAngleDistFunction = lambda self,pitchAngle_radians,rigidity:self.normFactor * np.exp((-0.5 * (pitchAngle_radians - (np.sin(pitchAngle_radians) * np.cos(pitchAngle_radians)))) / \
                                                                (self.A - (0.5 * (self.A - self.B) * (1-np.cos(pitchAngle_radians)))))

    def __init__(self,normFactor,A, B):
        self.normFactor = normFactor
        self.A = A
        self.B = B


import pandas as pd
import tqdm

import ParticleRigidityCalculationTools as PRCT

from pythonScripts.spectralCalculations.momentaDistribution import momentaDistribution

class WeightingFactorsSeries(pd.Series):
    @property
    def _constructor(self):
        return WeightingFactorsSeries

    @property
    def _constructor_expanddim(self):
        return WeightingFactorsDataFrame

class WeightingFactorsDataFrame(pd.DataFrame):

    @property
    def _constructor(self):
        return WeightingFactorsDataFrame

    @property
    def _constructor_sliced(self):
        return WeightingFactorsSeries

    def __init__(self,  AsymptoticDirectionDataframe, momentaDist:momentaDistribution): #formerly acquireWeightingFactors

        # initialise as empty dataframe
        super().__init__()

        # fill with values from original dataframe
        self.fillValuesFromDataFrame(AsymptoticDirectionDataframe)

        # find angle between asymptotic directions and IMF using dot product
        angleColumnTitle = "angleBetweenIMFinRadians"
        self.assignPitchAngles(momentaDist, angleColumnTitle)

        # find weighting factors from the angles and rigidities
        pitchAngleFunctionToUse = lambda row : momentaDist.getPitchAngleDistribution()(row[angleColumnTitle],row["Rigidity"])
        fullRigidityPitchWeightingFactorFunctionToUse = lambda row : momentaDist(row[angleColumnTitle],row["Rigidity"])

        self["PitchAngleWeightingFactor"] = self.apply(pitchAngleFunctionToUse,axis=1)
        self["RigidityWeightingFactor"] = self["Rigidity"].apply(momentaDist.getRigiditySpectrum()) * (self["Filter"] == 1)
        self["fullRigidityPitchWeightingFactor"] = self.apply(fullRigidityPitchWeightingFactorFunctionToUse,axis=1) * (self["Filter"] == 1)
        self["fullEnergyPitchWeightingFactor"] = PRCT.convertParticleRigiditySpecToEnergySpec(self["Rigidity"],self["fullRigidityPitchWeightingFactor"],
                                                                                       particleMassAU=1, particleChargeAU=1)

    def fillValuesFromDataFrame(self,dataframeToFillFrom:pd.DataFrame):

        self["Rigidity"] = dataframeToFillFrom["Rigidity"]
        self["Energy"] = PRCT.convertParticleRigidityToEnergy(dataframeToFillFrom["Rigidity"], particleMassAU = 1, particleChargeAU = 1)
        self["Filter"] = dataframeToFillFrom["Filter"]
        #print(spacepy.coordinates.Coords(dataframeToFillFrom.iloc[0]["Lat"],))

        print("assigning asymptotic coordinates")
        asymptoticDirectionList = []
        initialPositionList = []
        initialMomDirList = []
        for dataframeRow in tqdm.tqdm(dataframeToFillFrom.iterrows(),total=len(dataframeToFillFrom)):
          rowInSpacepyCoords = self.convertInitialCoordSetToSpacePy(dataframeRow[1], 100.0,"Lat","Long")
          asymptoticDirectionList.append(rowInSpacepyCoords)

        print("successfully converted asymptotic directions")

        self["initialLatitude"] = dataframeToFillFrom["initialLatitude"]
        self["initialLongitude"] = dataframeToFillFrom["initialLongitude"]
        self["initialZenith"] = dataframeToFillFrom["initialZenith"]
        self["initialAzimuth"] = dataframeToFillFrom["initialAzimuth"]

        self["Asymptotic Direction"] = asymptoticDirectionList


import os
import numpy as np

onekftinkm = 0.3048

class maireSrunManager():

    MAIREHOME = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    WORKDIR =MAIREHOME + "/developmentWorkDir/"
    #datetimeToRunAcross = dt.datetime(year=2010, month=5,day=24, hour=3, minute=20,second=34)
    _listOfDoseRateTypes = ["adose","edose","dosee","tn1","tn2","tn3","SEU","SEL"]

    generalIterativeInteger = 0

    def __init__(self):
        print("MAIREHOME is",self.MAIREHOME)

    def getMAIREHOME(self):
        return self.MAIREHOME

    def getWORKDIR(self):
        return self.WORKDIR

    def setRunMode(self, runMode):
        self._runMode = runMode

    def getKpIndex(self):
        return self._runMode

    def setDATETIMETORUNACROSS(self, datetimeToRunAcross):
        self._datetimeToRunAcross = datetimeToRunAcross

    def getDATETIMETORUNACROSS(self):
        return self._datetimeToRunAcross

    def setKpIndex(self, KpIndex):
        self._KpIndex = KpIndex

    def getKpIndex(self):
        return self._KpIndex

    def getGeneralIterativeInteger(self):

        intToReturn = self.generalIterativeInteger
        self.generalIterativeInteger = self.generalIterativeInteger + 1

        return intToReturn

    def setRigidityValues(self, rigidityValues:np.array):
        self._rigidityValues = rigidityValues

_maireSrunManager = maireSrunManager()


from setuptools import find_packages, setup
import os
import glob as gb

# get requirements for installation
lib_folder = os.path.dirname(os.path.realpath(__file__))
requirement_path = lib_folder + '/requirements.txt'
install_requires = []
if os.path.isfile(requirement_path):
    with open(requirement_path) as f:
        install_requires = f.read().splitlines()
        print("install requires:")
        print(install_requires)

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='AniMAIRE',
    packages=find_packages(exclude='pytests'),
    package_data={"AniMAIRE":[
                                "anisotropic_MAIRE_engine/data/*.csv"
                                         ]},
    version='1.0.10',
    description='Python library for running the anisotropic version of MAIRE+',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Chris S. W. Davis',
    author_email='ChrisSWDavis@gmail.com',
    license='GNU General Public License v3.0',
    url='https://github.com/ssc-maire/AniMAIRE-public',
    keywords='anisotropic MAIRE+ atmospheric ionizing radiation dose rates cosmic rays ground level enhancements GLEs protons alpha particles neutrons effective ambient equivalent aircraft aviation Earth solar system sun space magnetic field',
    install_requires=['numpy>=1.21.6',
                        'pandarallel>=1.6.3',
                        'pandas>=1.3.5',
                        'ParticleRigidityCalculationTools>=1.5.4',
                        'scipy>=1.7.3',
                        'setuptools>=45.2.0',
                        'spacepy>=0.3.0',
                        'tqdm>=4.65.0',
                        'plotly>=5.9.0',
                        'wheel>=0.41.2',
                        'numba>=0.57.1',
                        'metpy>=1.5.1',
                        'CosRayModifiedISO >= 1.2.3',
                        'AsympDirsCalculator >= 1.0.8',
                        'seaborn >= 0.13.2',
                        'geopandas >= 0.11.1',
                        'joblib >= 1.2.0',
                        'atmosphericRadiationDoseAndFlux >= 1.0.3'],
    #install_requires,
    setup_requires=['pytest-runner','wheel'],
    tests_require=['pytest'],
)



import numpy as np
import pandas as pd

from pandarallel import pandarallel
from .utils.AsymptoticDirectionDataframe import acquireWeightingFactors, get_apply_method
pandarallel.initialize(progress_bar=True)

from .spectralCalculations.particleDistribution import particleDistribution

from atmosphericRadiationDoseAndFlux import doseAndFluxCalculator as DAFcalc
from scipy.interpolate import interp1d
from scipy.integrate import trapz

import pkg_resources

import ParticleRigidityCalculationTools as PRCT

from numba import njit, vectorize, typed, types, typeof

NM64_proton_response = pd.read_csv(pkg_resources.resource_filename(__name__,'data/NM64_proton_response.csv'), 
                                    header=None)
NM64_proton_response.columns = ["proton_rigidity_GV","cts_per_primary_per_cm2"]

NM64_proton_filepath_Mishev_2013 = pkg_resources.resource_filename(__name__,'data/NM64_proton_response_Mishev_2013.csv')
NM64_proton_response_Mishev_2013_epn = pd.read_csv(NM64_proton_filepath_Mishev_2013,header=None)
NM64_proton_response_Mishev_2013_epn.columns = ["Energy_per_nucleon_GeV_per_n","Yield_arb_units"]
NM64_proton_response_Mishev_2013_epn["Yield_per_m2_per_sr"] = NM64_proton_response_Mishev_2013_epn["Yield_arb_units"] * 0.109
NM64_proton_response_Mishev_2013 = PRCT.convertParticleEnergySpecToRigiditySpec(particleKineticEnergyInMeV=NM64_proton_response_Mishev_2013_epn["Energy_per_nucleon_GeV_per_n"]*1000,
                                             fluxInEnergyMeVform=NM64_proton_response_Mishev_2013_epn["Yield_per_m2_per_sr"],
                                             particleMassAU=1,
                                             particleChargeAU=1)

NM64_alpha_filepath_Mishev_2013 = pkg_resources.resource_filename(__name__,'data/NM64_alpha_response_Mishev_2013.csv')
NM64_alpha_response_Mishev_2013_epn = pd.read_csv(NM64_alpha_filepath_Mishev_2013,header=None)
NM64_alpha_response_Mishev_2013_epn.columns = ["Energy_per_nucleon_GeV_per_n","Yield_arb_units"]
NM64_alpha_response_Mishev_2013_epn["Yield_per_m2_per_sr"] = NM64_alpha_response_Mishev_2013_epn["Yield_arb_units"] * 8.37e-2 * 4 #need to multiply by 4 because at the moment its in per nucleon
NM64_alpha_response_Mishev_2013 = PRCT.convertParticleEnergySpecToRigiditySpec(particleKineticEnergyInMeV=NM64_alpha_response_Mishev_2013_epn["Energy_per_nucleon_GeV_per_n"]*1000*4,
                                             fluxInEnergyMeVform=NM64_alpha_response_Mishev_2013_epn["Yield_per_m2_per_sr"],
                                             particleMassAU=4,
                                             particleChargeAU=2)

# 2020 response values and the response function below are taken from https://agupubs.onlinelibrary.wiley.com/doi/full/10.1029/2019JA027433

NM64_alpha_filepath_Mishev_2020 = pkg_resources.resource_filename(__name__,'data/NM64_alpha_response_Mishev_2020.csv')
NM64_alpha_response_Mishev_2020_epn = pd.read_csv(NM64_alpha_filepath_Mishev_2020,header=None)
NM64_alpha_response_Mishev_2020_epn.columns = ["Energy_per_nucleon_GeV_per_n","Yield_arb_units"]
NM64_alpha_response_Mishev_2020_epn["Yield_per_m2_per_sr"] = NM64_alpha_response_Mishev_2020_epn["Yield_arb_units"] * 4 #need to multiply by 4 because at the moment its in per nucleon
NM64_alpha_response_Mishev_2020 = PRCT.convertParticleEnergySpecToRigiditySpec(particleKineticEnergyInMeV=NM64_alpha_response_Mishev_2020_epn["Energy_per_nucleon_GeV_per_n"]*1000*4,
                                             fluxInEnergyMeVform=NM64_alpha_response_Mishev_2020_epn["Yield_per_m2_per_sr"],
                                             particleMassAU=4,
                                             particleChargeAU=2)

@njit
def convert_particle_energy_to_rigidity(atomic_mass, atomic_charge, particle_name, energy_GeV_per_nucleon):

    rest_mass_GeV_dict = {"proton":0.938,"alpha":3.727}

    R_value = (atomic_mass / atomic_charge) * np.sqrt(energy_GeV_per_nucleon * (energy_GeV_per_nucleon + (2 * rest_mass_GeV_dict[particle_name])))

    return R_value

# dict_of_parameter_values = typed.Dict.empty(
#                             key_type=types.unicode_type,
#                             value_type=types.float64[:],
#                         )
# dict_for_proton_limits = typed.Dict.empty(
#                         key_type=types.float64[:],
#                         value_type=typeof(dict_of_parameter_values),
#                     )

# dict_for_alpha_limits = typed.Dict.empty(
#                         key_type=types.float64[:],
#                         value_type=typeof(dict_of_parameter_values),
#                     )

# # dict_of_parameters = typed.Dict.empty(
# #                         key_type=types.unicode_type,
# #                         value_type=typeof(dict_for_limits),
# #                     )

# dict_for_proton_limits[0.0] = {"a0":-12.104,"a1":13.879,"a2":-8.6616,"a3":0.0,}

@np.vectorize
#@vectorize #(nopython=True)
def get_NM64_response_value(particle_name:str,energy_GeV_per_nucleon:float):
    
    dict_of_parameters={"proton":{
                                  0.0: {"a0":-12.104,"a1":13.879,"a2":-8.6616,"a3":0.0,},
                                  1.28:{"a0":-8.76,"a1":2.831,"a2":0.428,"a3":-0.186,},
                                  10.0:  {"a0":-4.763,"a1":1.206,"a2":-0.0365,"a3":0.0,}
                                  },
                        "alpha":{
                                  0.0: {"a0":-9.5396,"a1":7.2827,"a2":-3.0659,"a3":0.5082,},
                                  1.7:{"a0":-8.65,"a1":4.9329,"a2":-1.2022,"a3":0.1179,},
                                  15.0:  {"a0":-4.763,"a1":1.206,"a2":-0.0365,"a3":0.0,}
                                  }
                        }
    
    energy_limits_keys = list(dict_of_parameters[particle_name].keys())

    if energy_GeV_per_nucleon <= energy_limits_keys[1]:
        params_to_use = dict_of_parameters[particle_name][energy_limits_keys[0]]
    elif energy_GeV_per_nucleon < energy_limits_keys[2]:
        params_to_use = dict_of_parameters[particle_name][energy_limits_keys[1]]
    else:
        params_to_use = dict_of_parameters[particle_name][energy_limits_keys[2]]

    if particle_name == "proton":
        #R_value = PRCT.convertParticleEnergyToRigidity(energy_GeV_per_nucleon * 1000)
        R_value = convert_particle_energy_to_rigidity(1,1,"proton",energy_GeV_per_nucleon)
    elif particle_name == "alpha":
        #R_value = PRCT.convertParticleEnergyToRigidity(energy_GeV_per_nucleon * 1000 * 4, particleMassAU=4, particleChargeAU=2)
        R_value = convert_particle_energy_to_rigidity(4,2,"alpha",energy_GeV_per_nucleon)
    #R_value = np.sqrt(energy_GeV_per_nucleon * (energy_GeV_per_nucleon + 1.876))

    ln_yield_value = sum_yield_values(list(params_to_use.values()), R_value)

    return np.exp(ln_yield_value)

@njit
def sum_yield_values(params_to_use, R_value):
    ln_yield_value = 0.0
    for l in [0,1,2,3]:
        #ln_yield_value += params_to_use[f"a{l}"] * (np.log(R_value)**l)
        ln_yield_value += params_to_use[l] * (np.log(R_value)**l)
    return ln_yield_value

#@njit
def get_parameter_value(particle_name:str,parameter_label:str,energy_GeV_per_nucleon:float):

    dict_of_parameters={
                        "proton":{
                                  "A":{"b5":6.945e-9,"b4":-1.461e-7,"b3":1.115e-6,"b2":-3.402e-6,"b1":3.355e-6,"b0":-9.823e-7},
                                  "B":{"b5":-3.963e-6,"b4":8.091e-5,"b3":-6.394e-4,"b2":2.348e-3,"b1":-4.713e-3,"b0":1.186e-2}
                                 },
                        "alpha":{
                                  "A":{"b5":9.422e-9,"b4":-2.284e-7,"b3":2.037e-6,"b2":-7.828e-6,"b1":1.203e-5,"b0":-5.545e-6},
                                  "B":{"b5":-5.351e-6,"b4":1.316e-4,"b3":-1.226e-3,"b2":5.176e-3,"b1":-1.017e-2,"b0":1.458e-2}
                                 },
                        }

    params_to_use = dict_of_parameters[particle_name][parameter_label]
    
    if particle_name == "proton":
        #R_value = PRCT.convertParticleEnergyToRigidity(energy_GeV_per_nucleon * 1000)
        R_value = convert_particle_energy_to_rigidity(1,1,"proton",energy_GeV_per_nucleon)
    elif particle_name == "alpha":
        #R_value = PRCT.convertParticleEnergyToRigidity(energy_GeV_per_nucleon * 1000 * 4, particleMassAU=4, particleChargeAU=2)
        R_value = convert_particle_energy_to_rigidity(4,2,"alpha",energy_GeV_per_nucleon)
    #R_value = np.sqrt(energy_GeV_per_nucleon * (energy_GeV_per_nucleon + 1.876))

    output_parameter = 0.0
    for l in [0,1,2,3,4,5]:
        output_parameter += params_to_use[f"b{l}"] * (np.log(R_value)**l)

    return output_parameter

#@njit
def get_NM64_response_value_atmospheric_depth(particle_name:str,energy_GeV_per_nucleon:float,atmospheric_depth_g_per_cm2):

    ground_level_values = get_NM64_response_value(particle_name,energy_GeV_per_nucleon)

    A = get_parameter_value(particle_name, "A", energy_GeV_per_nucleon)
    B = get_parameter_value(particle_name, "B", energy_GeV_per_nucleon)

    values_at_depth = get_values_at_depth(atmospheric_depth_g_per_cm2, ground_level_values, A, B)

    return values_at_depth

@njit
def get_values_at_depth(atmospheric_depth_g_per_cm2, ground_level_values, A, B):
    values_at_depth = ground_level_values * np.exp((A * (1000-atmospheric_depth_g_per_cm2)**2) + (B * (1000-atmospheric_depth_g_per_cm2)) )
    return values_at_depth

from metpy.constants import g
from metpy.calc import height_to_pressure_std
from metpy.units import units

def get_NM64_response_value_altitude(particle_name:str,energy_GeV_per_nucleon:float,altitude_in_km:float):

    pressure_at_altitude_hectopascal = height_to_pressure_std(altitude_in_km * units("km"))
    atmospheric_depth_g_per_cm2 = (pressure_at_altitude_hectopascal / g).to("g / (cm**2)").magnitude

    values_at_altitude = get_NM64_response_value_atmospheric_depth(particle_name, energy_GeV_per_nucleon, atmospheric_depth_g_per_cm2)

    return values_at_altitude

#response_value_epns = list(range(0.1,1000,(1000-0.1)/1000))
#response_value_epns = np.geomspace(0.1,1000,100000)
response_value_epns = np.geomspace(0.1,1000,10000)

proton_response_values = get_NM64_response_value("proton",response_value_epns)

NM64_proton_response_Mishev_2020_functional_epn = pd.DataFrame({"Energy_per_nucleon_GeV_per_n":response_value_epns,
                                                                "Yield_per_m2_per_sr":proton_response_values})
NM64_proton_response_Mishev_2020_functional = PRCT.convertParticleEnergySpecToRigiditySpec(NM64_proton_response_Mishev_2020_functional_epn["Energy_per_nucleon_GeV_per_n"]*1000,
                                             fluxInEnergyMeVform=NM64_proton_response_Mishev_2020_functional_epn["Yield_per_m2_per_sr"],
                                             particleMassAU=1,
                                             particleChargeAU=1)

alpha_response_values = get_NM64_response_value("alpha",response_value_epns)

NM64_alpha_response_Mishev_2020_functional_epn = pd.DataFrame({"Energy_per_nucleon_GeV_per_n":response_value_epns,
              "Yield_per_m2_per_sr":alpha_response_values * 4})
NM64_alpha_response_Mishev_2020_functional = PRCT.convertParticleEnergySpecToRigiditySpec(NM64_alpha_response_Mishev_2020_functional_epn["Energy_per_nucleon_GeV_per_n"]*1000*4,
                                             fluxInEnergyMeVform=NM64_alpha_response_Mishev_2020_functional_epn["Yield_per_m2_per_sr"],
                                             particleMassAU=4,
                                             particleChargeAU=2)

class singleParticleEngineInstance():

    def __init__(self, 
                 particle_distribution:particleDistribution, 
                 dfofAsymptoticDirections:pd.DataFrame, 
                 list_of_altitudes_in_km:list,
                 generate_NM_count_rates:bool):
        
        self.particle_distribution = particle_distribution
        self.list_of_altitudes_in_km = list_of_altitudes_in_km
        self.dfOfAllAsymptoticDirections = dfofAsymptoticDirections
        self.generate_NM_count_rates=generate_NM_count_rates
        self.rigiditySpectrumParamDict = {}
        self.pitchAngleDistributionParamDict = {}

    def getThePitchAngleDistribution(self):
        print("acquiring pitch angle distributions...")
        pitchAngleDistToUse = self.particle_distribution.momentum_distribution.pitch_angle_distribution
        try:
            print("IMF latitude set to",self.IMFlatitude)
            print("IMF longitude set to",self.IMFlongitude)
        except AttributeError:
            self.IMFlatitude = -1000.0 # distribution is probably isotropic: while a better solution is probably necessary, set to dummy variables for now.
            self.IMFlongitude = -1000.0
        pitchAngleDistToUse.setInterplanetaryMagFieldDirection(self.IMFlatitude,self.IMFlongitude)
        return pitchAngleDistToUse

    def getAsymptoticDirsAndRun(self):

        self.acquireDFofAllAsymptoticDirections()
        sortedOutputDoseRates = self.runOverSpecifiedAltitudes()

        return sortedOutputDoseRates

    def runOverSpecifiedAltitudes(self):
        
        #######################################################
        # assign each asymptotic direction a weighting factor in accordance with its pitch angle
        print("assigning pitch angle weighting factors...")
        df_with_weighting_factors_full_angles = acquireWeightingFactors(self.dfOfAllAsymptoticDirections,
                                                            self.particle_distribution.momentum_distribution)
        #######################################################

        # print("df_with_weighting_factors:")
        # print(df_with_weighting_factors.to_string())
        # import pickle as pkl
        # with open("weightingFactorsDF_2.pkl","wb") as weightingFactorsFile:
        #     pkl.dump(df_with_weighting_factors, weightingFactorsFile)

        #######################################################
        df_with_weighting_factors_full_angles.to_pickle("df_with_weighting_factors_full_angles.pkl")
        # convert the dataframe such that the maxmimum weighting factor for the initial azimuths and zeniths is used.
        #df_with_weighting_factors = get_max_weighting_factors_for_multi_angle_magcos_runs(df_with_weighting_factors_full_angles)

        df_with_weighting_factors = get_mean_weighting_factors_for_multi_angle_magcos_runs(df_with_weighting_factors_full_angles)
        #######################################################

        #######################################################
        # calculate dose rates at each point based on spectra and weighting factors
        print("converting spectra and asymptotic directions to particle fluxes and dose rates...")

        sortedOutputDoseRates = self.calc_output_dose_flux(df_with_weighting_factors, 
                                                           self.list_of_altitudes_in_km, 
                                                           self.particle_distribution.particle_species.particleName)
        
        df_with_weighting_factors.to_pickle("weighting_factor_DF.pkl")
        sortedOutputDoseRates.weighting_factor_input_DF = df_with_weighting_factors
        
        print("output dose rates calculated successfully!")

        #exit()

        #raise Exception("mdsiondsidosdids")

        # 5 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        return sortedOutputDoseRates

    def calc_output_dose_flux(self, asymp_dir_DF_with_weighting_factors, list_of_altitudes_in_km, particle_name):

        spectrum_to_function_conversion_function = lambda row: interp1d(x=np.array(row["Rigidity"]),
                                                                        y=np.array(row["fullRigidityPitchWeightingFactor"]),
                                                                        bounds_error=False,
                                                                        fill_value=0.0)

        DFofSpectraForEachCoord = asymp_dir_DF_with_weighting_factors.groupby(["initialLatitude","initialLongitude"]).apply(spectrum_to_function_conversion_function)
        #print(DFofSpectraForEachCoord)
        outputDoseRatesForAltitudeRange = get_apply_method(DFofSpectraForEachCoord)(lambda spectrum:DAFcalc.calculate_from_rigidity_spec(
                                                                                                inputRigidityDistributionFunctionGV=lambda x:float(spectrum(x)),
                                                                                                altitudesInkm=list_of_altitudes_in_km,
                                                                                                particleName = particle_name,
                                                                                                )
                                                                            )
        
        #exit()
        outputDoseRatesOnlyDF = pd.concat(outputDoseRatesForAltitudeRange.tolist(), ignore_index=True)

        index_list = [-2,-1,0,1,2,3,4,5,6,7,8]

        if self.generate_NM_count_rates == True:
            print("calculating neutron monitor count rates...")
            outputDoseRatesOnlyDF["NM64_cr_unnorm"] = self.get_neutron_monitor_count_rates(list_of_altitudes_in_km, particle_name, DFofSpectraForEachCoord)
            print("neutron monitor count rates successfully determined.")

            index_list = [-2,-1,0,1,2,3,4,5,6,7,8,9]
        
        outputDoseRatesOnlyDF_with_lats_and_longs = self.reinsert_original_lats_and_longs(outputDoseRatesOnlyDF,outputDoseRatesForAltitudeRange)
        #######################################################

        sortedOutputDoseRates = outputDoseRatesOnlyDF_with_lats_and_longs  \
                                        .iloc[:,index_list] \
                                        .sort_values(["latitude","longitude","altitude (km)"], ignore_index=True)
        return sortedOutputDoseRates

    def get_neutron_monitor_count_rates(self, list_of_altitudes_in_km, particle_name, DFofSpectraForEachCoord):
        NM64_values_list = []

        for altitude_in_km in list_of_altitudes_in_km:
            NM64_vals_DF = get_apply_method(DFofSpectraForEachCoord)(lambda x:self.calculate_unnormed_NM64_cr(x,particle_name = particle_name, altitude_in_km=altitude_in_km)).values
            NM64_values_list.append(NM64_vals_DF)

        #outputDoseRatesOnlyDF["NM64_cr_unnorm"] = np.repeat(NM64_vals_DF,len(list_of_altitudes_in_km))
        new_cr_unnorm_list = np.array(NM64_values_list).T.flatten()
        return new_cr_unnorm_list

    def calculate_unnormed_NM64_cr(self, interpolated_spectra_row, particle_name, altitude_in_km = 0.0):
        #spectra_weighting_factors_across_NM = NM64_proton_response["proton_rigidity_GV"].apply(interpolated_spectra_row)

        #response_values = get_NM64_response_value(particle_name,response_value_epns)
        response_values = get_NM64_response_value_altitude(particle_name, response_value_epns, altitude_in_km)

        NM64_response_Mishev_2020_functional_epn = pd.DataFrame({"Energy_per_nucleon_GeV_per_n":response_value_epns,
                                                                        "Yield_per_m2_per_sr":response_values})

        if particle_name == "proton":
            # #NM64_response_to_use = NM64_proton_response_Mishev_2013
            # NM64_response_to_use = NM64_response_Mishev_2020_functional

            NM64_response_Mishev_2020_functional = PRCT.convertParticleEnergySpecToRigiditySpec(NM64_response_Mishev_2020_functional_epn["Energy_per_nucleon_GeV_per_n"]*1000,
                                                    fluxInEnergyMeVform=NM64_response_Mishev_2020_functional_epn["Yield_per_m2_per_sr"],
                                                    particleMassAU=1,
                                                    particleChargeAU=1)
            
        elif particle_name == "alpha":
            # #NM64_response_to_use = NM64_alpha_response_Mishev_2013
            # #NM64_response_to_use = NM64_alpha_response_Mishev_2020
            # NM64_response_to_use = NM64_alpha_response_Mishev_2020_functional
            # #NM64_response_to_use = NM64_proton_response_Mishev_2013
            # # NM64_response_to_use = pd.DataFrame({"Rigidity":[0.0,1.0],
            # #                                      "Rigidity distribution values":[0.0,0.0]})

            NM64_response_Mishev_2020_functional = PRCT.convertParticleEnergySpecToRigiditySpec(NM64_response_Mishev_2020_functional_epn["Energy_per_nucleon_GeV_per_n"] * 1000 * 4,
                                                    fluxInEnergyMeVform=NM64_response_Mishev_2020_functional_epn["Yield_per_m2_per_sr"],
                                                    particleMassAU=4,
                                                    particleChargeAU=2)

        else:
            raise Exception("ERROR: particle name did not match either proton or alpha!")
        
        NM64_response_to_use = NM64_response_Mishev_2020_functional

        spectra_weighting_factors_across_NM = NM64_response_to_use["Rigidity"].apply(interpolated_spectra_row)
        
        full_NM64_spectra_weighting_factor = spectra_weighting_factors_across_NM * NM64_response_to_use["Rigidity distribution values"]
        output_unnormalised_NM64_cr = trapz(full_NM64_spectra_weighting_factor, NM64_response_to_use["Rigidity"])
        return output_unnormalised_NM64_cr

    def reinsert_original_lats_and_longs(self, new_doserates_DF:pd.DataFrame,original_data_frame:pd.Series):
        original_lats = original_data_frame.reset_index(level=[0,1])["initialLatitude"]
        original_longs = original_data_frame.reset_index(level=[0,1])["initialLongitude"]

        list_of_altitude_dfs = []
        for altitude_in_km in new_doserates_DF["altitude (km)"].unique():
            df_for_specific_altitude = new_doserates_DF[new_doserates_DF["altitude (km)"] == altitude_in_km].reset_index(drop=True).copy()
            df_for_specific_altitude["latitude"] = original_lats
            df_for_specific_altitude["longitude"] = original_longs
            list_of_altitude_dfs.append(df_for_specific_altitude)

        return pd.concat(list_of_altitude_dfs, ignore_index=True)
    
def get_max_weighting_factors_for_multi_angle_magcos_runs(multi_angle_DF:pd.DataFrame):

    max_weighting_factor_DF = multi_angle_DF[multi_angle_DF[["Rigidity","initialLatitude","initialLongitude","fullRigidityPitchWeightingFactor"]].groupby(["Rigidity","initialLatitude","initialLongitude"], group_keys=False).apply(lambda x:x == x.max())["fullRigidityPitchWeightingFactor"]]

    return max_weighting_factor_DF.reset_index(drop=True)

def get_mean_weighting_factors_for_multi_angle_magcos_runs(multi_angle_DF:pd.DataFrame):

    max_weighting_factor_DF = multi_angle_DF.groupby(["Rigidity","initialLatitude","initialLongitude"], group_keys=False).mean()

    return max_weighting_factor_DF.reset_index()


import numpy as np
import pandas as pd
import sys

import ParticleRigidityCalculationTools as PRCT

from joblib import Memory

import tqdm
tqdm.tqdm.pandas()

from pandarallel import pandarallel
pandarallel.initialize(progress_bar=True)

from spacepy.coordinates import Coords as spaceCoords
from spacepy.time import Ticktock as spaceTicktock

import numba

memory_asymp_dirs = Memory("./cacheAsymptoticDirectionOutputs", verbose=0)

m0 = 1.67262192e-27 #kg
c = 299792458.0 #m/s
protonCharge = 1.60217663e-19 #C
protonRestEnergy = m0 * (c**2)

gettrace = getattr(sys, 'gettrace', None)
global get_apply_method
if gettrace is None:
    #print('No sys.gettrace')
    get_apply_method = lambda DF_or_Series:DF_or_Series.parallel_apply
else:
    get_apply_method = lambda DF_or_Series:DF_or_Series.progress_apply
    

def generate_asymp_dir_DF(dataframeToFillFrom:pd.DataFrame, IMFlatitude:float, IMFlongitude:float, datetime_to_run_across_UTC, cache:bool):

    new_asymp_dir_DF = dataframeToFillFrom.copy()

    new_asymp_dir_DF["Energy"] = PRCT.convertParticleRigidityToEnergy(dataframeToFillFrom["Rigidity"], particleMassAU = 1, particleChargeAU = 1)

    print("assigning asymptotic coordinates")
    if cache == False:
        asymptoticDirectionList = convertAsymptoticDirectionsToPitchAngle(dataframeToFillFrom, IMFlatitude, IMFlongitude, datetime_to_run_across_UTC)
    else:
        cachedConvertAsympDirFunc = memory_asymp_dirs.cache(convertAsymptoticDirectionsToPitchAngle)
        asymptoticDirectionList = cachedConvertAsympDirFunc(dataframeToFillFrom, IMFlatitude, IMFlongitude, datetime_to_run_across_UTC)

    print("successfully converted asymptotic directions")

    new_asymp_dir_DF["angleBetweenIMFinRadians"] = asymptoticDirectionList

    return new_asymp_dir_DF

#@memory_asymp_dirs.cache()
def convertInitialCoordSetToSpacePy(dataframeRow:pd.Series, radialPosition:float, firstCoordinateLabel:str, secondCoordinateLabel:str, datetime_to_run_across_UTC):

    rowLatitude = dataframeRow[firstCoordinateLabel]
    rowLongitude = dataframeRow[secondCoordinateLabel]

    rowInSpacepyCoords = spaceCoords([radialPosition,rowLatitude,rowLongitude],
                                     "GEO","sph",
                                     units=['Re', 'deg', 'deg'],
                                     use_irbem=False)
    rowInSpacepyCoords.ticks = spaceTicktock(datetime_to_run_across_UTC,"UTC")

    return rowInSpacepyCoords

def convertAsymptoticDirectionsToPitchAngle(dataframeToFillFrom, IMFlatitude, IMFlongitude, datetime_to_run_across_UTC):
    
    # pitch_angle_list = []
    # for index, dataframeRow in tqdm.tqdm(dataframeToFillFrom.iterrows(),total=len(dataframeToFillFrom)):
    #     pitch_angle_for_row = get_pitch_angle_for_DF_row(IMFlatitude, IMFlongitude, dataframeRow)
    #     pitch_angle_list.append(pitch_angle_for_row)

    print("acquiring pitch angles...")
    #print(dataframeToFillFrom)
    #pitch_angle_list = dataframeToFillFrom.parallel_apply(lambda dataframe_row:get_pitch_angle_for_DF_row(IMFlatitude, IMFlongitude, dataframe_row, datetime_to_run_across_UTC),axis=1)
    
    #pitch_angle_list = dataframeToFillFrom.parallel_apply(lambda dataframe_row:get_pitch_angle_for_DF_analytic(IMFlatitude, IMFlongitude, dataframe_row["Lat"], dataframe_row["Long"]),axis=1)
    pitch_angle_list = dataframeToFillFrom.progress_apply(lambda dataframe_row:get_pitch_angle_for_DF_analytic(IMFlatitude, IMFlongitude, dataframe_row["Lat"], dataframe_row["Long"]),axis=1)

    #pitch_angle_list = dataframeToFillFrom.parallel_apply(lambda dataframe_row:get_pitch_angle_for_DF_Mishev(IMFlatitude, IMFlongitude, dataframe_row["Lat"], dataframe_row["Long"]),axis=1)

    #print(pitch_angle_list)

    #asymptoticDirectionList = list(dataframeToFillFrom.progress_apply(lambda dataframeRow:convertInitialCoordSetToSpacePy(dataframeRow[1], 100.0,"Lat","Long")))

    # with pmp.Pool(processes = 4) as pmpPool:
    #     asymptoticDirectionList = pmpPool.map(lambda dataframeRow:convertInitialCoordSetToSpacePy(dataframeRow[1], 100.0,"Lat","Long"),
    #                                         tqdm.tqdm(list(dataframeToFillFrom.iterrows()),total=len(dataframeToFillFrom)))

    return pitch_angle_list

@numba.jit(nopython=True)
def get_pitch_angle_for_DF_analytic(IMFlatitude, IMFlongitude, asymptotic_dir_latitude, asymptotic_dir_longitude):

    IMFlatitude_rad = IMFlatitude * (np.pi/180.0)
    IMFlongitude_rad = IMFlongitude * (np.pi/180.0)
    asymptotic_dir_latitude_rad = asymptotic_dir_latitude * (np.pi/180.0)
    asymptotic_dir_longitude_rad = asymptotic_dir_longitude * (np.pi/180.0)

    cos_pitch_angle = (np.sin(asymptotic_dir_latitude_rad) * np.sin(IMFlatitude_rad)) + \
                      (np.cos(asymptotic_dir_latitude_rad) * np.cos(IMFlatitude_rad) * np.cos(asymptotic_dir_longitude_rad - IMFlongitude_rad))
    
    pitch_angle = np.arccos(cos_pitch_angle)

    return pitch_angle

def get_pitch_angle_for_DF_Mishev(IMFlatitude, IMFlongitude, asymptotic_dir_latitude, asymptotic_dir_longitude):

    IMFlatitude_rad = IMFlatitude * (np.pi/180.0)
    IMFlongitude_rad = IMFlongitude * (np.pi/180.0)
    asymptotic_dir_latitude_rad = asymptotic_dir_latitude * (np.pi/180.0)
    asymptotic_dir_longitude_rad = asymptotic_dir_longitude * (np.pi/180.0)

    cos_pitch_angle = (np.sin(asymptotic_dir_latitude_rad) * np.sin(IMFlatitude_rad)) + \
                      (np.cos(asymptotic_dir_latitude_rad) * np.cos(IMFlatitude_rad) * np.cos(IMFlongitude_rad) * (np.cos(asymptotic_dir_longitude_rad) + np.sin(asymptotic_dir_longitude_rad)))
    
    pitch_angle = np.arccos(cos_pitch_angle)

    return pitch_angle

def get_pitch_angle_for_DF_row(IMFlatitude, IMFlongitude, dataframeRow, datetime_to_run_across_UTC):
    #interplanetary_mag_field = spaceCoords([100.0,IMFlatitude, IMFlongitude],"GSM","sph",units=['Re', 'deg', 'deg'],use_irbem=False)
    interplanetary_mag_field = spaceCoords([100.0,IMFlatitude, IMFlongitude],"GEO","sph",units=['Re', 'deg', 'deg'],use_irbem=False)
    rowInSpacepyCoords = convertInitialCoordSetToSpacePy(dataframeRow, 100.0,"Lat","Long", datetime_to_run_across_UTC)
    pitch_angle_for_row = calculatePitchAngle_from_IMF_dir(interplanetary_mag_field, rowInSpacepyCoords, datetime_to_run_across_UTC)
    return pitch_angle_for_row

def acquireWeightingFactors(asymptotic_direction_DF:pd.DataFrame, momentaDist):

    new_asymptotic_direction_DF = asymptotic_direction_DF.copy()

    # # find angle between asymptotic directions and IMF using dot product
    # print("determining angles between asympotic directions and IMF...")
    # angleColumnTitle = "angleBetweenIMFinRadians"
    # new_asymptotic_direction_DF = assignPitchAngles(new_asymptotic_direction_DF, momentaDist, angleColumnTitle)

    # find weighting factors from the angles and rigidities
    #jacobian_function_to_use = lambda pitch_angle_in_radians:1/np.sin(2.0 * pitch_angle_in_radians)
        
    pitchAngleFunctionToUse = lambda row : momentaDist.getPitchAngleDistribution()(row["angleBetweenIMFinRadians"],row["Rigidity"])
    fullRigidityPitchWeightingFactorFunctionToUse = lambda row : momentaDist(row["angleBetweenIMFinRadians"],row["Rigidity"])

    #print(new_asymptotic_direction_DF)
    print("calculating pitch angle weighting factors...")
    #new_asymptotic_direction_DF["PitchAngleWeightingFactor"] = new_asymptotic_direction_DF.progress_apply(pitchAngleFunctionToUse,axis=1)
    #new_asymptotic_direction_DF["PitchAngleWeightingFactor"] = new_asymptotic_direction_DF.parallel_apply(pitchAngleFunctionToUse,axis=1)
    new_asymptotic_direction_DF["PitchAngleWeightingFactor"] = get_apply_method(new_asymptotic_direction_DF)(pitchAngleFunctionToUse,axis=1)
    #print("pitch angle weights:")
    #print(new_asymptotic_direction_DF[["angleBetweenIMFinRadians","PitchAngleWeightingFactor"]])
    #new_asymptotic_direction_DF["PitchAngleWeightingFactor"] = 1
    #print(new_asymptotic_direction_DF)

    #new_asymptotic_direction_DF["Filter"] = new_asymptotic_direction_DF["Filter"].loc[::-1].cummax().loc[::-1]
    #new_asymptotic_direction_DF = new_asymptotic_direction_DF[new_asymptotic_direction_DF["Filter"] == 1]
    new_asymptotic_direction_DF["Filter"] = (new_asymptotic_direction_DF["Filter"] == 1) * 1

    print("calculating rigidity weighting factors...")
    new_asymptotic_direction_DF["RigidityWeightingFactor"] = get_apply_method(new_asymptotic_direction_DF["Rigidity"])(momentaDist.getRigiditySpectrum())
    #print(new_asymptotic_direction_DF)

    print("calculating rigidity + pitch combined weighting factors...")
    new_asymptotic_direction_DF["fullRigidityPitchWeightingFactor"] = get_apply_method(new_asymptotic_direction_DF)(fullRigidityPitchWeightingFactorFunctionToUse,axis=1) * (new_asymptotic_direction_DF["Filter"] == 1)
    #print(new_asymptotic_direction_DF)

    print("calculating energy + pitch combined weighting factors...")
    energySpectrum = PRCT.convertParticleRigiditySpecToEnergySpec(new_asymptotic_direction_DF["Rigidity"],
                                                                  new_asymptotic_direction_DF["fullRigidityPitchWeightingFactor"],
                                                                  particleMassAU=1, particleChargeAU=1)

    new_asymptotic_direction_DF["fullEnergyPitchWeightingFactor"] = energySpectrum["Energy distribution values"]
    #print(new_asymptotic_direction_DF)

    return new_asymptotic_direction_DF

def assignPitchAngles(asymp_dir_DF:pd.DataFrame, momentaDist, angleColumnTitle):

    new_asymp_dir_DF = asymp_dir_DF.copy()
    
    # with pmp.Pool(processes = 8) as pmpPool:
    #     listOfOutputDoseDFs = pmpPool.map(lambda dfRow:self.calculatePitchAngle(momentaDist, dfRow),
    #                                         tqdm.tqdm(self.iterrows(),total=len(self)))

    angleArray = []
    print("WARNING: this loop could become slow for large dataframes. Optimising this dataframe could be done through pandas array manipulation or through numba or cpython")
    for index, dfRow in tqdm.tqdm(new_asymp_dir_DF.iterrows(),total=len(new_asymp_dir_DF)):
        angleBetweenIMF = calculatePitchAngle(momentaDist, dfRow)
        angleArray.append(angleBetweenIMF)
    new_asymp_dir_DF[angleColumnTitle] = angleArray

    #self[angleColumnTitle] = self.to_DataFrame()["Asymptotic Direction"].apply(lambda x:self.calculatePitchAngleApply(momentaDist,x))
    
    #pitchAngleDF = self.to_DataFrame()["Asymptotic Direction"].progress_apply(lambda x:self.calculatePitchAngleApply(momentaDist,x))
    #self[angleColumnTitle] = pitchAngleDF #for some reason I believe a dummy variable is REQUIRED - otherwise the series does not get set correctly. However for some reason this function is still not working.
    
    #self[angleColumnTitle] = self.to_DataFrame()["Asymptotic Direction"].parallel_apply(lambda x:self.calculatePitchAngleApply(momentaDist,x))

    return new_asymp_dir_DF

def calculatePitchAngle_from_IMF_dir(interplanetary_mag_field:spaceCoords, asymptotic_direction:spaceCoords, datetime_to_run_across_UTC):

    cartesianAsympDir = asymptotic_direction.convert("GEO","car").data[0]
    GEOIMF = interplanetary_mag_field
    GEOIMF.ticks = spaceTicktock(datetime_to_run_across_UTC,"UTC")
    cartesianIMF = GEOIMF.convert("GEO","car").data[0]
    return calculateAngleBetweenTheSpaceVectors(cartesianAsympDir, cartesianIMF)

def calculatePitchAngle(momentaDist, dfRow, datetime_to_run_across_UTC):

    cartesianAsympDir = dfRow["Asymptotic Direction"].convert("GEO","car").data[0]
    GEOIMF = momentaDist.getPitchAngleDistribution().interplanetary_mag_field
    GEOIMF.ticks = spaceTicktock(datetime_to_run_across_UTC,"UTC")
    cartesianIMF = GEOIMF.convert("GEO","car").data[0]
    return calculateAngleBetweenTheSpaceVectors(cartesianAsympDir, cartesianIMF)

#@staticmethod
#@jit(nopython=True)
def calculateAngleBetweenTheSpaceVectors(cartesianAsympDir, cartesianIMF):
    dotProduct = np.dot(cartesianAsympDir, cartesianIMF)
    amplitude = np.linalg.norm(cartesianAsympDir)*np.linalg.norm(cartesianIMF)
    angleBetweenIMF = np.arccos(dotProduct/amplitude)
    return angleBetweenIMF

def splitDataframeIntoUniqueCoordinates(self):

    columnsToSplitBy = ["initialLatitude","initialLongitude","initialZenith","initialAzimuth"]

    vanillaSelfDataFrame = self.to_DataFrame()

    uniqueRelevantRowsDF = vanillaSelfDataFrame[columnsToSplitBy].drop_duplicates()
    
    outputListOfSplitDFs = []
    for index, row in tqdm.tqdm(uniqueRelevantRowsDF.iterrows(),total=len(uniqueRelevantRowsDF)):
        useRowBoolean = True
        for column in columnsToSplitBy:
            useRowBoolean = (vanillaSelfDataFrame[column] == row[column]) & useRowBoolean
        
        outputListOfSplitDFs.append(vanillaSelfDataFrame[useRowBoolean])
    
    return outputListOfSplitDFs
